#include <../source/TAL/objbase.h>

#define SAFETY (vec_t)0.9
#define PGROW (vec_t)-0.2
#define PSHRNK (vec_t)-0.25
/* (5/SAFETY) ^ (1/PGROW) */
#define ERRCON (vec_t)1.89e-4
#define MAXSTP 10000

typedef struct cash_karp_context_s
{
	libv_object_t base;
	
	u32_t num_variables;

	/* user-supplied right side */
	void (V_CALL *derivs) (vec_t[], vec_t, vec_t[]);
	
	/* local storage for step (nr_rkck) routine */
	vec_t *ak2, *ak3, *ak4, *ak5, *ak6, *ytemp1;
	/* local storage for stepper (nr_rkqs) routine */
	vec_t *yerr, *ytemp2;
	/* local storage for driver (nr_odeint) routine */
	vec_t *yscal, *y, *dydx;
    /*
     * User storage for intermediate results.
     * 
     * kmax != 0 results are stored at approximate intervals
     * (x2-x1)/(kmax-1) in the arrays xp[0..kmax-1], yp[0..nvar][0..save_count],
     * where save_count is output by nr_odeint.
     */
    u32_t kmax;
    vec_t *xp, **yp;
    vector_p xpv;
    matrix_p ypmat;
    vboolean verbose;
    u32_t max_steps;
    /* counts for 'good' and 'bad' steps */
    u32_t nok, nbad;
}	cash_karp_context_t, *cash_karp_context_p;

#ifdef	CHECKED_BUILD
#define CKCTX_MAGIC	0x434B4354 /* CKCT */
V_LOCAL char_p ckctx_name = "RK5-CK ODE integration object";
#endif	/* CHECKED_BUILD */

/*
 * Management of the storage for intermediate results
 */
V_LOCAL void
ck_free_istorage(cash_karp_context_p ctx)
{
	ctx->kmax = 0;
	ctx->xp = NULL;
	ctx->yp = NULL;
	if(ctx->xpv)
		memFree(ctx->xpv);
	if(ctx->ypmat)
		memFree(ctx->ypmat);
	ctx->xpv = NULL;
	ctx->ypmat = NULL;
}

V_LOCAL void ck_alloc_istorage(cash_karp_context_p ctx,u32_t num_saves)
{
	if(ctx->kmax)
		ck_free_istorage(ctx);
    if(num_saves)
    {
    	ctx->xpv = vecAlloc(num_saves);
    	ctx->ypmat = matAlloc(ctx->num_variables,num_saves);
    	ctx->xp = ctx->xpv->elements;
    	ctx->yp = ctx->ypmat->cell;
    	ctx->kmax = num_saves;
    }
}

/*
 * Destructor
 */
V_LOCAL u32_t V_CALL
destroy_ck_context(vhandle_t handle)
{
	cash_karp_context_p self = (cash_karp_context_p)handle;
	if(--self->base.ref_count)
		return self->base.ref_count;
	ck_free_istorage(self);
	memFreeObject(self);
	return 0;
}

/*
 * Constructor
 */	
V_LOCAL cash_karp_context_p create_ck_context(u32_t num_variables,
	void (V_CALL *derivs) (vec_t[], vec_t, vec_t[]),
	u32_t num_saves,vboolean verbose)
{
	cash_karp_context_p ctx;
	ALLOC_OBJECT(ctx,cash_karp_context_t,
		num_variables * 11 * sizeof(vec_t),CKCTX_MAGIC);

	ctx->base.destroy=destroy_ck_context;

    ctx->num_variables = num_variables;
	ctx->derivs = derivs;
	
	ctx->ak2 = (vec_p)(ctx+1) - 1; /* NR uses unit - based arrays */
	ctx->ak3 = ctx->ak2 + num_variables;
	ctx->ak4 = ctx->ak3 + num_variables;
	ctx->ak5 = ctx->ak4 + num_variables;
	ctx->ak6 = ctx->ak5 + num_variables;
	ctx->ytemp1 = ctx->ak6 + num_variables;

	ctx->yerr = ctx->ytemp1 + num_variables;
	ctx->ytemp2 = ctx->yerr + num_variables;

	ctx->yscal = ctx->ytemp2 + num_variables;
	ctx->y = ctx->yscal + num_variables;
	ctx->dydx = ctx->y + num_variables;

	ctx->max_steps = MAXSTP;
	/* ctx->kmax == 0 now, so next line will not chain to ck_free_istorage */
	ck_alloc_istorage(ctx,num_saves);

	ctx->verbose = verbose;
    return ctx;
}

/*
 * Given values for n variables y[1..n] and their derivatives
 * dydx[1..n] known at x, use the fifth-order Cash-Karp Runge-Kutta
 * method to advance the solution over an interval h and return the
 * incremented variables as yout[1..n]. Also return an estimate of
 * the local truncation error in yout using the embedded fourth-order
 * method. The user supplies the routine derivs(dydx,x,y), which
 * returns derivatives dydx at x.
 */

V_LOCAL void
nr_rkck(cash_karp_context_p context,
	vec_t y[],
	vec_t dydx[],
	vec_t x,
	vec_t h,
	vec_t yout[],
	vec_t yerr[])
{
	const vec_t a2 = (vec_t)0.2, a3 = (vec_t)0.3, a4 = (vec_t)0.6,
    	a5 = (vec_t)1.0, a6 = (vec_t)0.875;
	const vec_t b21 = (vec_t)0.2, b31 = (vec_t)(3.0 / 40.0),
		b32 = (vec_t)(9.0 / 40.0), b41 = (vec_t)0.3, b42 = (vec_t)-0.9,
		b43 = (vec_t)1.2, b51 = (vec_t)(-11.0 / 54.0), b52 = (vec_t)2.5,
		b53 = (vec_t)(-70.0 / 27.0), b54 = (vec_t)(35.0 / 27.0),
		b61 = (vec_t)(1631.0 / 55296.0), b62 = (vec_t)(175.0 / 512.0),
		b63 = (vec_t)(575.0 / 13824.0), b64 = (vec_t)(44275.0 / 110592.0),
		b65 = (vec_t)(253.0 / 4096.0);
	const vec_t c1 = (vec_t)(37.0 / 378.0), c3 = (vec_t)(250.0 / 621.0),
		c4 = (vec_t)(125.0 / 594.0), c6 = (vec_t)(512.0 / 1771.0),
		dc5 = (vec_t)(-277.00 / 14336.0);
	const vec_t dc1 = (vec_t)(37.0 / 378.0) - (vec_t)(2825.0 / 27648.0);
	const vec_t dc3 = (vec_t)(250.0 / 621.0) - (vec_t)(18575.0 / 48384.0);
	const vec_t dc4 = (vec_t)(125.0 / 594.0) - (vec_t)(13525.0 / 55296.0);
	const vec_t dc6 = (vec_t)(512.0 / 1771.0) - (vec_t)0.25;

	void (V_CALL *derivs) (vec_t[], vec_t, vec_t[]) = context->derivs;

	u32_t i, n = context->num_variables;


	vec_t *ak2, *ak3, *ak4, *ak5, *ak6, *ytemp;

	ak2 = context->ak2;
	ak3 = context->ak3;
	ak4 = context->ak4;
	ak5 = context->ak5;
	ak6 = context->ak6;
	ytemp = context->ytemp1;

	for (i = 1; i <= n; i++)    /* First step. */
		ytemp[i] = y[i] + b21 * h * dydx[i];
	(*derivs) (ak2+1, x + a2 * h, ytemp+1); /* Second step. */
	for (i = 1; i <= n; i++)
		ytemp[i] = y[i] + h * (b31 * dydx[i] + b32 * ak2[i]);
	(*derivs) (ak3+1, x + a3 * h, ytemp+1); /* Third step. */
	for (i = 1; i <= n; i++)
		ytemp[i] = y[i] + h * (b41 * dydx[i] + b42 * ak2[i] + b43 * ak3[i]);
	(*derivs) (ak4+1, x + a4 * h, ytemp+1); /* Fourth step. */
	for (i = 1; i <= n; i++)
		ytemp[i] = y[i] + h * (b51 * dydx[i] + b52 * ak2[i] + b53 * ak3[i] +
			b54 * ak4[i]);
	(*derivs) (ak5+1, x + a5 * h, ytemp+1); /* Fifth step. */
	for (i = 1; i <= n; i++)
		ytemp[i] = y[i] + h * (b61 * dydx[i] + b62 * ak2[i] + b63 * ak3[i] +
			b64 * ak4[i] + b65 * ak5[i]);
	(*derivs) (ak6+1, x + a6 * h, ytemp+1); /* Sixth step. */
	for (i = 1; i <= n; i++)    /* Accumulate increments with proper weights. */
		yout[i] = y[i] + h * (c1 * dydx[i] + c3 * ak3[i] + c4 * ak4[i] +
			c6 * ak6[i]);
	/* Estimate error as difference between fourth and fifth order methods. */
	for (i = 1; i <= n; i++)
		yerr[i] = h * (dc1 * dydx[i] + dc3 * ak3[i] + dc4 * ak4[i] +
			dc5 * ak5[i] + dc6 * ak6[i]);
}

/*
 * Fifth-order Runge-Kutta step with monitoring of local truncation error
 * to ensure accuracy and adjust stepsize. Input are the dependent
 * variable vector y[1..n] and its derivative dydx[1..n] at the
 * starting value of the independent variable x. Also input are
 * the stepsize to be attempted htry, the required accuracy eps,
 * and the vector yscal[1..n] against which the error is scaled.
 * On output, y and x are replaced by their new values, hdid is
 * the stepsize that was actually accomplished, and hnext is the
 * estimated next stepsize.
 * derivs is the user-supplied routine that computes the right-hand
 * side derivatives.
 */

#define FMAX(op1,op2)	(op1 > op2 ? op1 : op2)
#define FMIN(op1,op2)	(op1 > op2 ? op2 : op1)

V_LOCAL vboolean
nr_rkqs(cash_karp_context_p context,
	vec_t y[],
    vec_t dydx[],
    vec_t *x,
    vec_t htry,
    vec_t eps,
	vec_t yscal[],
    vec_t *hdid,
    vec_t *hnext)
{
	u32_t i, n = context->num_variables;
	vec_t yerr_by_yscal, errmax, h, htemp, xnew, *yerr, *ytemp;

	yerr = context->yerr;
	ytemp = context->ytemp2;

    h = htry; /* Set stepsize to the initial trial value. */
	for (;;) {
		nr_rkck(context, y, dydx, *x, h, ytemp, yerr);   /* Take a step. */
		errmax = (vec_t)0.0;           /* Evaluate accuracy. */
		for (i = 1; i <= n; i++) {
			yerr_by_yscal = (vec_t)v_abs(yerr[i] / yscal[i]);
			errmax = FMAX(errmax,yerr_by_yscal);
		}
		errmax /= eps;          /* Scale relative to required tolerance. */
		if (errmax <= 1.0)
			break;              /* Step succeeded. Compute size of next step. */
		htemp = SAFETY * h * (vec_t)pow(errmax, PSHRNK);
		/* Truncation error too large, reduce stepsize. */
		{
			vec_t h01 = (vec_t)0.1 * h;
			h = (h >= 0.0 ? FMAX(htemp, h01) : FMIN(htemp, h01));
		}
		/* No more than a factor of 10. */
		xnew = (*x) + h;
		if (xnew == *x)
		{
			if(context->verbose)
				conlPuts(LOG_ERROR,"nr_rkqs(): stepsize underflow.\n");
			return false;
		}
	}
	if (errmax > ERRCON)
		*hnext = SAFETY * h * (vec_t)pow(errmax, PGROW);
	else
		*hnext = h * (vec_t)5;       /* No more than a factor of 5 increase. */
	*x += (*hdid = h);
	for (i = 1; i <= n; i++)
		y[i] = ytemp[i];
	return true;
}

/*
 * Runge-Kutta driver with adaptive stepsize control. Integrate starting
 * values ystart[1..nvar] from x1 to x2 with accuracy eps, storing
 * intermediate results in global variables. h1 should be set as a
 * guessed first stepsize, hmin as the minimum allowed stepsize
 * (can be zero). On output nok and nbad are the number of good and
 * bad (but retried and fixed) steps taken, and ystart is replaced
 * by values at the end of the integration interval. derivs is the
 * user-supplied routine for calculating the right-hand side derivative,
 * while rkqs is the name of the stepper routine to be used.
 */

#define TINY (vec_t)1.0e-30

V_LOCAL vboolean
nr_odeint(cash_karp_context_p context,
	vec_t ystart[],
    vec_t x1,
    vec_t x2,
    vec_t eps,
    vec_t h1,
	vec_t hmin)
{
	u32_t nstp, i, nvar = context->num_variables;
    u32_t kmax = context->kmax;
    u32_t save_count = 0;
    vec_t *xp = context->xp, **yp = context->yp;
	vec_t dxsav, xsav, x, hnext, hdid, h;
	vec_t *yscal, *y, *dydx;
    /* u32_t *nok = &context->nok, *nbad = &context->nbad; */

	void (V_CALL *derivs) (vec_t[], vec_t, vec_t[]) = context->derivs;

	yscal = context->yscal;
	y = context->y;
	dydx = context->dydx;

	x = x1;
	h = x2 > x1 ? h1 : -h1;
	*nok = (*nbad) = 0;
	for (i = 1; i <= nvar; i++)
		y[i] = ystart[i];
	if (kmax) {
		dxsav = (kmax > 1) ? (x2-x1) / (kmax-1) : (x2-x1);
		xsav = x - dxsav * (vec_t)2.0; /* Assures storage of first step. */
	} else {
		/* make compiler happy */
		xsav = dxsav = 0;
	}
	for (nstp = 1; nstp <= context->max_steps; nstp++) {    /* Take at most MAXSTP steps. */
		(*derivs)(dydx+1, x, y+1);
		for (i = 1; i <= nvar; i++)
			/* Scaling used to monitor accuracy. 
			 * This general-purpose choice can be modified if need be. */
			yscal[i] = v_abs(y[i]) + v_abs(dydx[i] * h) + TINY;
		if (save_count < kmax && v_abs(x - xsav) > v_abs(dxsav)) {
			/* printf("save_count %d, x %g, y[0] %g\n",save_count,x,y[1]); */
			/*
			if(context->progress)
				(*context->progress)(nstp,x,y+1);
			*/
			
			
			xp[save_count] = x;    /* Store intermediate results. */
			
			for (i = 0; i < nvar; i++)
				yp[i][save_count] = y[i+1];
			save_count++;
			xsav = x;
		}
		if ((x + h - x2) * (x + h - x1) > 0.0)
			h = x2 - x;         /* If stepsize can overshoot, decrease. */
		if(!(*nr_rkqs)(context,y, dydx, &x, h, eps, yscal, &hdid, &hnext))
			return false;
		/*
		if(IsEqualEpsilon(hdid,h,ERRCON))
			++(*nok);
		else
			++(*nbad);
		*/
		if ((x - x2) * (x2 - x1) >= 0.0) {      /* Are we done? */
			for (i = 1; i <= nvar; i++)
				ystart[i] = y[i];
			if (kmax) {
				if(save_count >= kmax)
					save_count = kmax-1;
				xp[save_count] = x;        /* Save final step. */
				for (i = 0; i < nvar; i++)
					yp[i][save_count] = y[i+1];
			}
			return true;
		}
		if (v_abs(hnext) <= hmin)
		{
			if(context->verbose)
				conlPuts(LOG_ERROR,"nr_odeint(): Step size too small.\n");
			return false;
		}
		h = hnext;
	}
	if(context->verbose)
		conlPuts(LOG_ERROR,"nr_odeint(): Too many steps.\n");
	return false;
}

/*
 * Member functions
 */

vhandle_t V_GLOBAL
odeCreateSolver(u32_t num_variables,void (V_CALL *derivs)(vec_t dest_dxdy[],vec_t x,vec_t y[]))
{
	cash_karp_context_p ctx = create_ck_context(num_variables,derivs,100,false);
	return (vhandle_t)ctx;
}

void	V_GLOBAL
odeDestroySolver(vhandle_t handle)
{
	libvCloseHandle(handle,"odeDestroySolver()",ckctx_name,CKCTX_MAGIC);
}

vboolean V_GLOBAL
odeIntegrate(vhandle_t handle,vec_t ystart[],vec_t x1,vec_t x2,vec_t eps,vec_t h1,vec_t hmin)
{
	cash_karp_context_p ctx = (cash_karp_context_p)libvIsHandle(handle,
		"odeIntegrate()",ckctx_name,CKCTX_MAGIC);
	if(ctx)
		return nr_odeint(ctx,ystart-1,x1,x2,eps,h1,hmin);
	return false;
}


/*
 * DEBUG
 */

void V_CALL deriv(vec_t dest[],vec_t x,vec_t y[])
{
	dest[0] = y[0] * exp(x);
}

int main()
{
	vhandle_t solver;
	vec_t y;
	
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	solver = odeCreateSolver(1,deriv);
	if(!solver)
		return -1;
	
	y = 1;
	odeIntegrate(solver,&y,0,2,1e-10,1e-6,1e-30);
	
	printf("Y now %g\n",y);
	
	odeDestroySolver(solver);
	
	return 0;
}
